# Copyright 2023 S&P Global Commodity Insights

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#       http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class AuthError(Exception):
    """
    [401, 403] - Invalid Username, Password or Appkey
    """

    pass


class PerSecondLimitError(Exception):
    """
    [429] - Too Many Requests Per Second
    """

    pass


class DailyLimitError(Exception):
    """
    [429] - Too Many Requests Per Day
    """

    pass



# Copyright 2023 S&P Global Commodity Insights

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#       http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Union, Any, TypeVar, Collection, Optional
from pandas import Series
from typing_extensions import TypeGuard
from enum import Enum
from datetime import date

T = TypeVar("T", bound=Enum)


def list_to_filter(
    field_name: str,
    items: Optional[
        Union[
            Collection[str],
            Collection[float],
            Collection[T],
            Collection[date],
            Collection[bool],
            str,
            bool,
            date,
            float,
            T,
        ]
    ] = None,
    strategy: str = "platts",
) -> str:
    single_quote = "'"
    double_quote = '"'
    delim = ":" if strategy == "platts" else " eq"
    quote = f"{double_quote if strategy == 'platts' else single_quote}"
    if isinstance(items, Series):
        return convert_series_to_filterexp(field_name, items, quote)

    if not isinstance(items, (bool)) and not items:
        return ""

    if isinstance(items, bool):
        return f"{field_name}{delim} {items}"
    if isinstance(items, str):
        return f"{field_name}{delim} {quote}{items}{quote}"
    if isinstance(items, Enum):
        return f"{field_name}{delim} {quote}{items.value}{quote}"
    if isinstance(items, float) or isinstance(items, int):
        return f"{field_name}{delim} {items}"
    if isinstance(items, date):
        if strategy == "odata":
            return f"{field_name}{delim} {items}"
        else:
            return f"{field_name}{delim} {quote}{items}{quote}"

    if is_enum_list(items):
        n_items: List[str] = [x.value for x in items]
        n_items = [f"{quote}{x}{quote}" for x in n_items]
    elif is_str_list(items):
        n_items = [f"{quote}{x}{quote}" for x in items]
    elif is_bool_list(items):
        n_items = [f"{x}" for x in items]
    elif is_int_list(items):
        n_items = [str(x) for x in items]
    elif is_date_list(items):
        # if strategy == "odata":
        #     n_items = [str(x) for x in items]
        # else:
        n_items = [f"{quote}{x}{quote}" for x in items]
    else:
        raise TypeError("not supported")

    return f"{field_name} in ({','.join(n_items)})"


def odata_list_to_filter(
    field_name: str,
    items: Optional[
        Union[
            Collection[str],
            Collection[float],
            Collection[T],
            Collection[date],
            str,
            float,
            date,
            T,
        ]
    ] = None,
) -> str:
    return list_to_filter(field_name, items, strategy="odata")


def convert_series_to_filterexp(field_name: str, s: "Series[Any]", quote: str) -> str:
    if is_int_list(s):
        l = [str(x) for x in s]
    elif is_str_list(s):
        l = [f"{quote}{x}{quote}" for x in s]
    elif is_date_list(s):
        l = [f"{quote}{x}{quote}" for x in s]
    else:
        raise TypeError("series type unsupported")

    return f"{field_name} in ({','.join(l)})"


def convert_date_to_filter_exp(
    field_name: str,
    gt_param: Union[date, None],
    gte_param: Union[date, None],
    lt_param: Union[date, None],
    lte_param: Union[date, None],
    filter_params: List[str],
) -> List[str]:
    if gt_param is not None:
        filter_params.append(f'{field_name} > "{gt_param}"')
    if gte_param is not None:
        filter_params.append(f'{field_name} >= "{gte_param}"')
    if lt_param is not None:
        filter_params.append(f'{field_name} < "{lt_param}"')
    if lte_param is not None:
        filter_params.append(f'{field_name} <= "{lte_param}"')

    return filter_params


def is_enum_list(lst: Collection[Any]) -> TypeGuard[List[Enum]]:
    return all(isinstance(x, Enum) for x in lst)


def is_str_list(lst: Collection[Any]) -> TypeGuard[List[str]]:
    return all(isinstance(x, str) for x in lst)


def is_bool_list(lst: Collection[Any]) -> TypeGuard[List[bool]]:
    return all(isinstance(x, bool) for x in lst)


def is_date_list(lst: Collection[Any]) -> TypeGuard[List[date]]:
    return all(isinstance(x, date) for x in lst)


def is_int_list(lst: Collection[Any]) -> TypeGuard[List[float]]:
    return all(isinstance(x, float) for x in lst) or all(
        isinstance(x, int) for x in lst
    )



# Copyright 2023 S&P Global Commodity Insights

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#       http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
S&P Global Commodity Insights API\n
More details available here: https://developer.platts.com

Getting Started
---------------
>>> import spgci as ci
>>> ci.set_credentials(``username``, ``password``)
>>> ci.MarketData().get_assessments_by_symbol(symbol="PCAAS00")
\n
"""

from .market_data import MarketData
from .forward_curves import ForwardCurves
from .ewindow_md import EWindowMarketData
from .wos import WorldOilSupply
from .wrd import WorldRefineryData
from .oil_demand import GlobalOilDemand
from .api_client import get_token
from .epf import EnergyPriceForecast
from .insights import Insights
from .na_gas import NANaturalGasAnalytics
from .crude_supply_risk import CrudeAnalytics
from .giem import GlobalIntegratedEnergyModel
from .lng_analytics import LNGGlobalAnalytics
from .arbflow import Arbflow
from .structured_heards import StructuredHeards
from .weather import Weather
from .chemicals import Chemicals
from .eu_gas_analytics import EUGasAnalytics
from .oil_ngl_analytics import OilNGLAnalytics


from .config import username, password, set_credentials, version

__version__ = version


__all__ = [
    "MarketData",
    "username",
    "password",
    "set_credentials",
    "ForwardCurves",
    "EWindowMarketData",
    "WorldOilSupply",
    "WorldRefineryData",
    "get_token",
    "EnergyPriceForecast",
    "Insights",
    "LNGGlobalAnalytics",
    "GlobalOilDemand",
    "NANaturalGasAnalytics",
    "CrudeAnalytics",
    "Arbflow",
    "GlobalIntegratedEnergyModel",
    "StructuredHeards",
    "Weather",
    "Chemicals",
    "EUGasAnalytics",
    "OilNGLAnalytics",
]



# Copyright 2023 S&P Global Commodity Insights

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#       http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from requests import Response
from pandas import DataFrame, Series
import pandas as pd
from typing import Union, Optional, List
from spgci.api_client import get_data, Paginator
from spgci.utilities import list_to_filter
from datetime import date
from typing_extensions import Literal


class CrudeAnalytics:
    """
    Short-term supply-side price risk index. This index captures the price influence of changes in oil supply in the 0-30 day outlook period.
    Factors are assessed on the country level: A positive score indicates a bullish supply impact on price, whereas a negative score indicates a bearish impact.
    The absolute value of the score is an indicator of the magnitude of the impact.
    The global price risk index is an aggregate of country-level risk scores, weighted by the country’s share of global crude and condensate capacity.

    Includes
    --------
    ``get_country_scores()`` get oil supply risk country scores.
    ``get_country_total_scores()`` get aggregated country scores.

    """

    _path = "crude-analytics/v1/oil-supply-risk/"

    @staticmethod
    def _convert_to_df(resp: Response) -> pd.DataFrame:
        j = resp.json()
        df = pd.json_normalize(j["results"])  # type: ignore

        if "scoring_date" in df.columns:
            df["scoring_date"] = pd.to_datetime(df["scoring_date"])  # type: ignore

        return df

    @staticmethod
    def _paginate(resp: Response) -> Paginator:
        j = resp.json()
        count = j["metaData"]["count"]
        size = j["metaData"]["pageSize"]

        remainder = count % size
        quotient = count // size
        total_pages = quotient + (1 if remainder > 0 else 0)

        if total_pages <= 1:
            return Paginator(False, "page", 1)

        return Paginator(True, "page", total_pages=total_pages)

    def get_country_scores(
        self,
        *,
        scoring_date: Optional[Union[date, list[date], "Series[date]"]] = None,
        scoring_date_gt: Optional[date] = None,
        scoring_date_gte: Optional[date] = None,
        scoring_date_lt: Optional[date] = None,
        scoring_date_lte: Optional[date] = None,
        status: Optional[Literal["Current", "Historical"]] = None,
        region: Optional[Union[str, list[str], "Series[str]"]] = None,
        country: Optional[Union[str, list[str], "Series[str]"]] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        Analytics data for North America Natural Gas Pipelines .

        Parameters
        ----------

        scoring_date : Optional[Union[date, list[date], Series[date]]], optional
            filter by scoring_date, by default None
        scoring_date_gt : Optional[date], optional
            filter by ``scoring_date > x``, by default None
        scoring_date_gte : Optional[date], optional
            filter by ``scoring_date >= x``, by default None
        scoring_date_lt : Optional[date], optional
            filter by ``scoring_date < x``, by default None
        scoring_date_lte : Optional[date], optional
            filter by ``scoring_date <= x``, by default None
        status : Optional[Literal['Current', 'Historical']], optional
            filter by status, by default None
        region : Optional[Union[str, list[str], Series[str]]], optional
            filter by region, by default None
        country : Optional[Union[str, list[str], Series[str]]], optional
            filter by country, by default None
        filter_exp : Optional[str], optional
            pass-thru ``filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to request a particular page of results, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
              DataFrame
                  DataFrame of the ``response.json()``
              Response
                  Raw ``requests.Response`` object
        Examples
        --------
        **Get all scores for country**
        >>> ci.CrudeAnalytics().get_country_scores(country="United States")

        **Get latest scores for all countries**
        >>> ci.CrudeAnalytics().get_country_scores(status="Current")
        """
        endpoint_path = "country-scores"

        filter_param: List[str] = []

        filter_param.append(list_to_filter("scoringDate", scoring_date))
        filter_param.append(list_to_filter("status", status))
        filter_param.append(list_to_filter("region", region))
        filter_param.append(list_to_filter("country", country))

        if scoring_date_gt:
            filter_param.append(f'scoringDate > "{scoring_date_gt}"')
        if scoring_date_gte:
            filter_param.append(f'scoringDate >= "{scoring_date_gte}"')
        if scoring_date_lt:
            filter_param.append(f'scoringDate < "{scoring_date_lt}"')
        if scoring_date_lte:
            filter_param.append(f'scoringDate <= "{scoring_date_lte}"')

        filter_param = [fp for fp in filter_param if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_param)
        elif len(filter_param) > 0:
            filter_exp = " AND ".join(filter_param) + " AND (" + filter_exp + ")"

        params = {
            "pageSize": page_size,
            "filter": filter_exp,
            "page": page,
        }
        return get_data(
            path=f"{self._path}{endpoint_path}",
            df_fn=self._convert_to_df,
            params=params,
            paginate=paginate,
            paginate_fn=self._paginate,
            raw=raw,
        )

    def get_country_total_scores(
        self,
        *,
        status: Optional[Literal["Current", "Historical"]] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        Analytics data for North America Natural Gas Pipelines .

        Parameters
        ----------

        status : Optional[Literal['Current', 'Historical']], optional
            filter by status, by default None
        filter_exp : Optional[str], optional
            pass-thru ``filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to request a particular page of results, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
              DataFrame
                  DataFrame of the ``response.json()``
              Response
                  Raw ``requests.Response`` object
        Examples
        --------
        **Simple Usage**
        >>> ci.CrudeAnalytics().get_country_total_scores()
        """
        endpoint_path = "country-total-scores"

        filter_param: List[str] = []

        filter_param.append(list_to_filter("status", status))
        filter_param = [fp for fp in filter_param if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_param)
        elif len(filter_param) > 0:
            filter_exp = " AND ".join(filter_param) + " AND (" + filter_exp + ")"

        params = {
            "pageSize": page_size,
            "filter": filter_exp,
            "page": page,
        }
        return get_data(
            path=f"{self._path}{endpoint_path}",
            df_fn=self._convert_to_df,
            params=params,
            paginate=paginate,
            paginate_fn=self._paginate,
            raw=raw,
        )
